#include "config.h"
#include <inttypes.h>
#include <str91x.h>
#include <pl011-uart.h>
#include <UART.hh>
#include <convert.h>

//-----------------------------------------------------------------------------

//
// Could be called in static constructor phase, w/o ordering,
// so do not depend on other modules.
//
void
UARTbase::init(int no, uint8_t *rxp, int rxsz, uint8_t *txp, int txsz)
{
    uart = &UART(no);

    uart->cr   = 0;
    uart->imsc = 0;
    clear();
    rate(9600);
    parity(None);
}

void
UARTbase::clear()
{
    unsigned state = disable();

    uart->rsr = 0;	// clear errors

    restore(state);
}

void
UARTbase::start()
{
    disable();

    uart->cr = UART_CR_RXE | UART_CR_TXE | UART_CR_ENABLE;
}

//-----------------------------------------------------------------------------

unsigned
UARTbase::disable()
{
    unsigned cr = uart->cr;
    if (cr & UART_CR_ENABLE) {
	uart->cr = cr & ~UART_CR_ENABLE;	// disable UART
	while (uart->fr & UART_FR_BUSY)		// wait until transmitter idle
	    ;
	unsigned lcr = uart->lcr;		// flush transmit FIFO
	uart->lcr = lcr & ~UART_LCR_FEN;
	uart->lcr = lcr;
    }

    return cr & UART_CR_ENABLE;
}

void
UARTbase::restore(unsigned state)
{
    uart->cr |= state;
}

//-----------------------------------------------------------------------------

void
UARTbase::rate(unsigned rate)
{
    unsigned state = disable();

    unsigned divisor = ((__Config_BRCLK * (2*64/16)) / rate) + 32;
    divisor = (divisor + divisor % 2) / 2;
    uart->ibrd = divisor / 64;
    uart->fbrd = divisor % 64;
    uart->lcr = uart->lcr; // commit

    restore(state);
}

void
UARTbase::parity(int parity)
{
    unsigned state = disable();

    static const uint8_t line_control[5] =
	{ UART_LCR_FEN | UART_LCR_WLEN8 | UART_LCR_STP1,
	  UART_LCR_FEN | UART_LCR_WLEN8 | UART_LCR_STP1 | UART_LCR_PEN | UART_LCR_EVEN,
	  UART_LCR_FEN | UART_LCR_WLEN8 | UART_LCR_STP1 | UART_LCR_PEN | UART_LCR_ODD,
	  UART_LCR_FEN | UART_LCR_WLEN8 | UART_LCR_STP1 | UART_LCR_PEN | UART_LCR_SPS | UART_LCR_EVEN,
	  UART_LCR_FEN | UART_LCR_WLEN8 | UART_LCR_STP1 | UART_LCR_PEN | UART_LCR_SPS | UART_LCR_ODD,
	};
    uart->lcr = line_control[parity];

    restore(state);
}

//-----------------------------------------------------------------------------

void
UARTbase::set_break()
{
    unsigned state = disable();
    uart->lcr |= UART_LCR_BRK;
    restore(state);
}

void
UARTbase::clear_break()
{
    unsigned state = disable();
    uart->lcr &= ~UART_LCR_BRK;
    restore(state);
}

//-----------------------------------------------------------------------------

void
UARTbase::put(char c)
{
    {
	while (uart->fr & UART_FR_TXFF)
	    ;
	uart->dr = c;
    }
}

void
UARTbase::puts(const char *s)
{
    while (*s)
	put(*s++);
}

void
UARTbase::flush()
{
    while (!(uart->fr & UART_FR_TXFE))
	;
    while (uart->fr & UART_FR_BUSY)
	;
}

//-----------------------------------------------------------------------------

static int
xlat_error(unsigned data)
{
    if (data & UART_DR_BE)
	return UARTbase::Break;
    if (data & UART_DR_FE)
	return UARTbase::Framing;
    if (data & UART_DR_PE)
	return UARTbase::Parity;
    if (data & UART_DR_OE)
	return UARTbase::Overrun;
    return 0;
}

int
UARTbase::get(bool wait)
{
    {
	while (uart->fr & UART_FR_RXFE) {
	    if (!wait)
		return Empty;
	}
	unsigned c = uart->dr;
	if (c & ~UART_DR_DATA)
	    return xlat_error(c);
	return c & UART_DR_DATA;
    }
}

//-----------------------------------------------------------------------------

void 
UARTbase::putu(unsigned value, int ndigits, int radix)
{
    char temp[12];
    if (ndigits > 10)
	ndigits = 10;
    puts(ucvt(temp, value, ndigits, radix));
}

void 
UARTbase::puti(int value, int ndigits)
{
    char temp[12];
    if (ndigits > 10)
	ndigits = 10;
    puts(icvt(temp, value, ndigits));
}

