#ifndef __ARM_IRQ_H__
#define __ARM_IRQ_H__

#define PSR_N		(1<<31)
#define PSR_Z		(1<<30)
#define PSR_C		(1<<29)
#define PSR_V		(1<<28)
#define PSR_IRQ		(1<<7)
#define PSR_FIQ		(1<<6)
#define PSR_Thumb	(1<<5)
#define PSR_Mode_Mask	0x1F
#define PSR_Mode_User	0x10
#define PSR_Mode_FIQ	0x11
#define PSR_Mode_IRQ	0x12
#define PSR_Mode_Super	0x13
#define PSR_Mode_Abort	0x17
#define PSR_Mode_Undef	0x1B
#define PSR_Mode_System	0x1F

/*
 * In my setup, there is a common IRQ handler for all IRQs. This handler
 * saves at least the volatile context, then calls the final interrupt 
 * handler. Afterwards, the interrupt is acknowledged and the context 
 * restored.
 *
 * Do not disable both IRQ and FIQ in the same cycle to avoid surprise
 * interrupts. IRQ should be disabled when FIQ is disabled (assumption
 * by interrupt handler).
 * 
 * The single-tasking version allows nested IRQs, running on the SYS stack.
 * The AKX version offers two implementations. If nested interrupts are
 * allowed, interrupts run on the SVC stack, otherwise on the IRQ stack.
 */

//-----------------------------------------------------------------------------
#ifdef __thumb__

  #ifdef __cplusplus
  extern "C" {
  #endif

      // startup.S:
      unsigned IntEnable     (void);		// IRQ
      unsigned IntDisable    (void);		// IRQ
      unsigned IntEnableAll  (void);		// FIQ+IRQ
      unsigned IntDisableAll (void);		// FIQ+IRQ
      void     IntRestore    (unsigned saved);

  #ifdef __cplusplus
  }
  #endif

  #undef FIQ_FUNCTION				// impossible in thumb mode
  #define IRQ_FUNCTION				// see ARM for comment
  #define IRQ_VECTOR(x)	((uint32_t)(x) | 1)	// function to vector

//-----------------------------------------------------------------------------
#else

  // In AKX, calling IntDisable from within Critical creates a stack frame, 
  // so inlines are actually shorter.

  static __inline unsigned IntEnable ()
  {
      unsigned saved, temp;
      asm volatile ("mrs	%0, cpsr" : "=r"(saved));
      asm volatile ("bic	%0, %1, %2" : "=r"(temp) : "r"(saved), "I"(PSR_IRQ));
      asm volatile ("msr	cpsr_c, %0" : : "r"(temp));
      return saved;
  }
  static __inline unsigned IntDisable ()
  {
      unsigned saved, temp;
      asm volatile ("mrs	%0, cpsr" : "=r"(saved));
      asm volatile ("orr	%0, %1, %2" : "=r"(temp) : "r"(saved), "I"(PSR_IRQ));
      asm volatile ("msr	cpsr_c, %0" : : "r"(temp));
      return saved;
  }
  static __inline unsigned IntEnableAll ()
  {
      unsigned saved, temp;
      asm volatile ("mrs	%0, cpsr" : "=r"(saved));
      asm volatile ("bic	%0, %1, %2" : "=r"(temp) : "r"(saved), "I"(PSR_IRQ|PSR_FIQ));
      asm volatile ("msr	cpsr_c, %0" : : "r"(temp));
      return saved;
  }
  static __inline unsigned IntDisableAll ()
  {
      unsigned saved, temp;
      asm volatile ("mrs	%0, cpsr" : "=r"(saved));
      asm volatile ("orr	%0, %1, %2" : "=r"(temp) : "r"(saved), "I"(PSR_IRQ));
      asm volatile ("msr	cpsr_c, %0" : : "r"(temp));
      asm volatile ("orr	%0, %0, %2" : "=r"(temp) : "r"(saved), "I"(PSR_FIQ));
      asm volatile ("msr	cpsr_c, %0" : : "r"(temp));
      return saved;
  }
  static __inline void IntRestore (unsigned saved)
  {
      asm volatile ("msr	cpsr_c, %0" : : "rn"(saved));
  }

  #define FIQ_FUNCTION __attribute__((interrupt("FIQ")))
  #define IRQ_FUNCTION
  #define IRQ_VECTOR(x)	((uint32_t)(x))		// function to vector

#endif
//-----------------------------------------------------------------------------

#endif
