// STR91x ARM startup script.
// Inspired by James P. Lynch and some sample code.
// Supports thumb mode IRQ vector support.
// Defines weak default exception handlers.


//-----------------------------------------------------------------------------
// Application specific definitions
//-----------------------------------------------------------------------------

#define OrderedTCM	0	// => errata 1.16
#define BufferedAHB	1

//
// Define the size of various stacks.
// Note that there is no user mode stack defined.
//

#ifndef UND_STACK_SIZE
 #define UND_STACK_SIZE	4	// stack size for "undefined instruction" mode
#endif

#ifndef ABT_STACK_SIZE
 #define ABT_STACK_SIZE	4	// stack size for "abort" mode
#endif

#ifndef FIQ_STACK_SIZE
 #define FIQ_STACK_SIZE	4	// stack size for "FIQ" mode
#endif

#ifndef IRQ_STACK_SIZE
 // 9 words per nested interrupt, support 8 levels of nesting
 #define IRQ_STACK_SIZE	(8*9*4)	// stack size for "IRQ" mode
#endif

#ifndef SVC_STACK_SIZE
 #define SVC_STACK_SIZE	4	// stack size for "supervisor" mode
#endif

// Overridable exception handlers.
// The defaults just loop forever.
// __UNDF_handler:	undefined instruction handler
// __SWI_handler:	software interrupt handler
// __PABT_handler:	program abort handler
// __DABT_handler:	data abort handler
// __FIQ_handler:	FIQ handler

// Overridable initialization.
// __PLL_init:		setup PLL before C++ constructors are called
//					default does nothing.

//-----------------------------------------------------------------------------
// Macros
//-----------------------------------------------------------------------------

// begin a function
	.macro	beginFcn name
	.global	\name
	.func	\name
\name:
	.endm
	
// end a function
	.macro	endFcn name
	.size	\name, . - \name
	.endfunc
	.endm
	
// switch from Thumb to ARM mode
	.macro	toarm
	.thumb
	adr	r3, 9f
	bx	r3
	.arm
	.align	2
9:
	.endm
	
// begin a function which may be called in both modes
// optional 2nd argument when both entry points are required
#ifdef __thumb__
	.macro	thumbFcn thumbname,armname
	.thumb_func
beginFcn \thumbname
	toarm
	.ifnc \armname,
beginFcn \armname
	.endif
	.endm
#else
	.macro	thumbFcn thumbname,armname
beginFcn \thumbname
	.endm
#endif

//-----------------------------------------------------------------------------
// Interrupt definitions
//-----------------------------------------------------------------------------

	.arm

	.set	VIC0_BASE,   0xFFFFFF00			// => MVN
	.set	VIC0_VECTOR, (0xFFFFF030-VIC0_BASE)	// VIC0 vector register
	.set	VIC1_BASE,   0xFC000000			// => MOV
	.set	VIC1_VECTOR, (0xFC000030-VIC1_BASE)	// VIC1 vector register

	.set	MODE_USR, 0x10		// User Mode
	.set	MODE_FIQ, 0x11		// FIQ Mode
	.set	MODE_IRQ, 0x12		// IRQ Mode
	.set	MODE_SVC, 0x13		// Supervisor Mode
	.set	MODE_ABT, 0x17		// Abort Mode
	.set	MODE_UND, 0x1B		// Undefined Mode
	.set	MODE_SYS, 0x1F		// System Mode

	.equ	I_BIT, 0x80		// when I bit is set, IRQ is disabled
	.equ	F_BIT, 0x40		// when F bit is set, FIQ is disabled

	.set	SCU_SCR0, 0x5C002034
	.set	FMI_BASE, 0x54000000
	.set	FMI_BBSR,   0x00
	.set	FMI_NBBSR,  0x04
	.set	FMI_BBADR,  0x0C
	.set	FMI_NBBADR, 0x10
	.set	FMI_CR,     0x18
	.set	FMI_SR,     0x1C
	
	.global	main			// int main(void)
	.global	__sys_init		// void __sys_init(void)

	.global	_etext			// -> .data initial values in ROM
	.global	_data			// -> .data area in RAM
	.global	_edata			// end of .data area
	.global	__bss_start		// -> .bss area in RAM
	.global	__bss_end__		// end of .bss area

	.global	_stacks			// top of stack, see machine specific linker script
	.global	_bbank_addr
	.global	_bbank_size_code
	.global	_nbbank_addr
	.global	_nbbank_size_code
	.global	_ram_size_code

//-----------------------------------------------------------------------------
// Address 0, vectors
//-----------------------------------------------------------------------------
	.section .boot,"ax",%progbits

	.global	__reset
	.global	__start
	.global	__boot
	.func	__boot
__boot:

//
// Runtime Interrupt Vectors.
//
_vectors:	ldr	pc, RESET_addr		// 00 = RESET
		ldr	pc, UND_addr		// 04 = undefined instruction
		ldr	pc, SWI_addr		// 08 = software interrupt
		ldr	pc, PABT_addr		// 0C = program abort
		ldr	pc, DABT_addr		// 10 = data abort
		nop				// 14 = reserved
		ldr	pc, IRQ_addr		// 18 = IRQ
		ldr	pc, FIQ_addr		// 1C = FIQ

RESET_addr:	.word	__start			// 20
UND_addr:	.word	__UND_handler		// 24
SWI_addr:	.word	__SWI_handler		// 28
PABT_addr:	.word	__PABT_handler		// 2C
DABT_addr:	.word	__DABT_handler		// 30
IRQ_addr:	.word	__IRQ_handler		// 34
FIQ_addr:	.word	__FIQ_handler		// 38
__reset:	ldr	pc, RESET_addr		// 3C

		.size __boot, . - __boot
		.endfunc

		.text

//-----------------------------------------------------------------------------
// Startup code
//-----------------------------------------------------------------------------

// Setup the operating mode & stack.
// ---------------------------------
	.global	__start
	.global	_exit
	.func	__start
__start:

	nop				// wait for OSC stabilization
	nop
	nop
	nop
	nop
	nop
	nop
	nop

	mov	r0, #2*1024*1024	// wait for JTAG
1:	subs	r0, r0, #1
	bne	1b

// Basic hardware initializations.
// -----------------------------------------------

__init:	
#if OrderedTCM
	// ARM966E-S configuration control register, initially 0x00000000
	mov	r0, #(1<<17)|(1<<18)	// ITCM,DTCM ordered mode
	mcr	p15, 1, r0, c15, c1, 0
#else	
	mov	r0, #(0<<17)|(0<<18)	// ITCM,DTCM unordered mode
	mcr	p15, 1, r0, c15, c1, 0
#endif

#if BufferedAHB
	// ARM966E-S control register, initially 0x00001F74
	//	15=0:	LDPC loads T
	//	13=0:	low vectors
	//	12=1:	ITCM enabled
	//	 7=0:	little endian
	//	 3=0:	AHB unbuffered
	//	 2=1:	DTCM enabled
	//	 1=0:	address alignment fault checking disabled
	mrc	p15, 0, r0, c1, c0, 0   // 
	orr	r0, r0, #1<<3		// enable AHB write buffer
	mcr	p15, 0, r0, c1, c0, 0
#endif
	
	mov	r0, #FMI_BASE

	// FMI already initialized?
//	ldr     r1, [r0, #FMI_BBSR]	// boot bank size
//	ldr	r2, [r0, #FMI_NBBSR]	// non-boot bank size
//	adds	r1, r1, r2		// power-up clears both
//	bne	1f

	// no, power-up, map flash banks and enable
	ldr	r1, =_bbank_size_code	// boot bank size
        str	r1, [r0, #FMI_BBSR]
        ldr	r1, =_nbbank_size_code	// non-boot bank size
        str	r1, [r0, #FMI_NBBSR]
	ldr	r1, =_bbank_addr	// boot bank base address
	mov	r1, r1, lsr #2
        str	r1, [r0, #FMI_BBADR]
	ldr	r1, =_nbbank_addr	// non-boot bank base address
	mov	r1, r1, lsr #2
        str	r1, [r0, #FMI_NBBADR]
1:
	mov	r1, #0x18		// flash write 1WS, out of mem int disabled
	str	r1, [r0, #FMI_CR]	// flash bank int disabled, both banks enabled
	mov	r1, #0x03		// clear bank errors
	str	r1, [r0, #FMI_SR]
	
	ldr	r0, =SCU_SCR0		// PFQ/BC disabled, arbiter lock disabled
	ldr	r1, =0x186		// 1 AHB wait state, 1 DTCM wait state
	ldr	r2, =_ram_size_code	// N*32KB SRAM
	orr	r1, r1, r2, lsl #3
	str	r1, [r0]

// Initialize Interrupt System
// - Set stack location for each mode
// - Leave in System Mode with Interrupts Disabled
// -----------------------------------------------
	ldr	r0, =_stacks
	msr	CPSR_c, #MODE_UND|I_BIT|F_BIT	// Undefined Instruction Mode
	mov	sp, r0
	sub	r0, r0, #UND_STACK_SIZE
	msr	CPSR_c, #MODE_ABT|I_BIT|F_BIT	// Abort Mode
	mov	sp, r0
	sub	r0, r0, #ABT_STACK_SIZE
	msr	CPSR_c, #MODE_FIQ|I_BIT|F_BIT	// FIQ Mode
	mov	sp, r0
	sub	r0, r0, #FIQ_STACK_SIZE
	msr	CPSR_c, #MODE_IRQ|I_BIT|F_BIT	// IRQ Mode
	mov	sp, r0
	sub	r0, r0, #IRQ_STACK_SIZE
	msr	CPSR_c, #MODE_SVC|I_BIT|F_BIT	// Supervisor Mode
	mov	sp, r0
	sub	r0, r0, #SVC_STACK_SIZE
	msr	CPSR_c, #MODE_SYS|I_BIT|F_BIT	// System Mode
	mov	sp, r0

// Copy initialized data to its execution address in RAM
// -----------------------------------------------------
	ldr	r1, =_idata			// -> ROM data start
	ldr	r2, =_data			// -> data start
	ldr	r3, =_edata			// -> end of data
1:      cmp	r2, r3				// check if data to move
	ldrlo	r0, [r1],#4			// copy it
	strlo	r0, [r2],#4
	blo	1b				// loop until done

// Clear .bss
// ----------
	mov	r0, #0				// get a zero
	ldr	r1, =__bss_start		// -> bss start
	ldr	r2, =__bss_end__		// -> bss end
2:      cmp	r1, r2				// check if data to clear
	strlo	r0, [r1],#4			// clear 4 bytes
	blo	2b				// loop until done

// Call C++ constructors (for objects in "global scope")
// System initialization must come before, or units are in reset.
// -----------------------------------------------------

	adr	lr, 0f
	ldr	pc, =__sys_init
0:
	ldr	r4, =__ctors_start__
	ldr	r5, =__ctors_end__
1:	cmp	r4, r5
	adrne	lr, 1b
	ldrne	pc, [r4], #4
	
// Call main program: main(0)
// --------------------------
	mov	r0, #0				// no arguments (argc = 0)
	mov	r1, #0
	mov	r2, #0
	mov	fp, #0				// null frame pointer
	mov	r7, #0				// null frame pointer for thumb
	adr	lr, _exit
	ldr	pc, =main
_exit:	b	.

//!!	.pool	//Mll im bin-file bei thumb
	
// "global object"-dtors are never called and it should not be 
// needed since there is no OS to exit to.

	.size	__start, . - __start
	.endfunc

//-----------------------------------------------------------------------------
// IRQ handler.
//-----------------------------------------------------------------------------
	.weak	__IRQ_handler
	.func	__IRQ_handler
__IRQ_handler:
	sub	lr, lr, #4			// fix return address
	stmfd	sp!, {lr}			// save return address
	mrs	r14, spsr			// save workreg and SPSR_IRQ
	stmfd	sp!, {r14}

	msr	cpsr_c, #MODE_SYS|I_BIT		// switch to system mode, IRQ disabled, FIQ enabled

	stmfd	sp!, {r0-r3,r12,lr}		// save caller saved registers (LR_SYS)
	mvn	r0, #~VIC0_BASE
	ldr	r0, [r0,#VIC0_VECTOR]		// load/freeze VIC0 interrupt vector
	mov	r1, #VIC1_VECTOR
	ldr	r1, [r1,#VIC1_VECTOR]		// freeze VIC1 interrupt vector
	blx	r0
	ldmfd	sp!, {r0-r3,r12,lr}		// restore caller saved registers

	msr	cpsr_c, #MODE_IRQ|I_BIT		// back to IRQ stack, IRQ disabled, FIQ enabled
	
	mvn	r0, #~VIC0_BASE
	str	r0, [r0,#VIC0_VECTOR]		// acknowledge VIC0 interrupt
	mov	r1, #VIC1_BASE
	str	r1, [r1,#VIC1_VECTOR]		// acknowledge VIC1 interrupt

	ldmfd	sp!, {r14}			// restore workreg and SPSR_IRQ
	msr	spsr, r14
	ldmfd	sp!, {pc}^			// return and restore CPSR from SPSR_IRQ

	.pool

	.size	__IRQ_handler, . - __IRQ_handler
	.endfunc

//-----------------------------------------------------------------------------
// Default exception handlers.
//-----------------------------------------------------------------------------
	.macro	default_handler name
	.weak	\name
	.func	\name
\name:	b	.
	.size	\name, . - \name
	.endfunc
	.endm

	default_handler	__UND_handler
	default_handler	__SWI_handler
	default_handler	__PABT_handler
	default_handler	__DABT_handler
	default_handler	__FIQ_handler

//-----------------------------------------------------------------------------
// Enable/Disable interrupts. Do not disable both interrupts in the same cycle, 
// to avoid surprise interrupts.
//-----------------------------------------------------------------------------

#ifdef __thumb__

thumbFcn IntEnable
	mrs	r0, cpsr
	bic	r1, r0, #I_BIT
	msr	cpsr_c, r1
	bx	lr
endFcn IntEnable

thumbFcn IntDisable
	mrs	r0, cpsr
	orr	r1, r0, #I_BIT
	msr	cpsr_c, r1
	bx	lr
endFcn IntDisable

thumbFcn IntDisableAll
	mrs	r0, cpsr
	orr	r1, r0, #I_BIT
	msr	cpsr_c, r1
	orr	r1, r1, #F_BIT
	msr	cpsr_c, r1
	bx	lr
endFcn IntDisable

thumbFcn IntRestore
	msr	cpsr_c, r0
	bx	lr
endFcn IntRestore

#endif

//-----------------------------------------------------------------------------
// Access core control registers.
//-----------------------------------------------------------------------------

#ifdef __thumb__

thumbFcn arm9_get_idcode
	mrc	p15, 0, r0, c0, c0, 0
	bx	lr
endFcn arm9_get_idcode

thumbFcn arm9_get_tcmsize
	mrc	p15, 0, r0, c0, c0, 2
	bx	lr
endFcn arm9_get_tcmsize

thumbFcn arm9_get_control
	mrc	p15, 0, r0, c1, c0, 0
	bx	lr
endFcn arm9_get_control

thumbFcn arm9_set_control
	mcr	p15, 0, r0, c1, c0, 0
	bx	lr
endFcn arm9_set_control

thumbFcn arm9_wait_for_irq
	mrc	p15, 0, r0, c7, c0, 4
	bx	lr
endFcn arm9_wait_for_irq

thumbFcn arm9_drain_buffers
	mcr	p15, 0, r0, c7, c10, 4
	bx	lr
endFcn arm9_drain_buffers
	
thumbFcn arm9_get_config
	mrc	p15, 0, r0, c1, c0, 0
	bx	lr
endFcn arm9_get_config

thumbFcn arm9_set_config
	mcr	p15, 0, r0, c1, c0, 0
	bx	lr
endFcn arm9_set_config

#endif

	.end
